# Copyright (C) 2019  Chair of Automatic Control, Kiel University
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from firedrake.petsc import PETSc
from firedrake import *
import slepc4py
import os
import sys
import numpy as np
import scipy as sp

from mpi4py import MPI

comm = MPI.COMM_WORLD


class dyn_solver:
    '''Class for the transient solution of second order systems.

    This class uses the SS22 solver described in:

    O. C. Zienkiewicz, R. L. Taylor, J. Z. Zhu,
    The Finite Element Method: Its Basis and Fundamentals,
    6th Edition, Butterworth-Heinemann, 2005

    System representation:
    M ddx + C dx + K x + f = 0
    with f = B u

    Attributes:
    M: System matrix (PETSc matrix)
    C: System matrix (PETSc matrix)
    K: System matrix (PETSc matrix)
    B: System matrix/vector (PETSc matrix/vector)
    T: End time of the simulation
    dt: Stepsize of the simulation
    N: Number of steps to take for the solver
    ksp_tran: Linear system solver for the transient solution PETSc.KSP
    solver_parameter: fix parameter for the solver
    A: Fix left hand side matrix (see [Zienkiewicz]) (PETSc matrix)
    t_sol: Solution time as list
    x_sol: Solution states, list of PETSc vectors

    Args:
    M: System matrix (PETSc matrix)
    C: System matrix (PETSc matrix)
    K: System matrix (PETSc matrix)
    B: System matrix/vector (PETSc matrix/vector)
    End_Time: End time for dynamic solution
    delta_t: Timestep
    system_string: Optional describtion of the system, e.g. 'full system model'
    '''

    def __init__(self,
                 M,
                 C,
                 K,
                 B,
                 end_time,
                 delta_t,
                 system_string=''):
        '''Init function'''

        self.M = M
        self.C = C
        self.K = K
        self.B = B

        self.T = end_time
        self.dt = delta_t

        if system_string != '':
            self._sys_str = system_string
        else:
            self._sys_str = ''

        self.ksp_tran = PETSc.KSP()
        self.ksp_tran.create()

        self.solver_parameter = [
            0.5, 0.5
        ]  # Values from Zienkiewicz Page 606 ( Explanation of SS22)

        self.A = self.M + self.solver_parameter[
            1] * self.dt * self.C + 0.5 * self.solver_parameter[0] * pow(
                self.dt, 2) * self.K

        self.ksp_tran.setOperators(self.A)

    def start_solve(self, ini_val, input_func, inputfunc_args=''):
        '''Solving function

        Returns the solution for each timestep!

        Args:
        ini_val : Intial value has the form [ x(t=t0) , dx(t=t0)  ]
        input_func : Funtion for the input! This function has to have the form: input_func(t,x) and has to return the input at the aktuall timestep!
        inputfunc_args : Arguments for input function

        Returns:
        The solution in form of a list [t_sol,x_sol].
        '''

        if self._sys_str == '':
            tmp = '\tDynamic solution:'
        else:
            tmp = '\tDynamic solution for ' + self._sys_str + ':'

        print(tmp)

        self.N = (self.T / self.dt)

        if self.N % 1 != 0:
            self.N = int(np.ceil(self.N))
            self.T = self.N * self.dt
            print(
                '\t\t-The Simulation time T_End isn\'t matching with the size of timesteps. The endtime was raised to {}'
                .format(self.T))
        else:
            self.N = int(self.N)

        self.x_sol = []
        self.t_sol = []

        print('\t\tStart of dynamic solution')

        self._SS22(ini_val, input_func, inputfunc_args)

        print('\t\tFinished dynamic solution!')

        return [self.t_sol, self.x_sol]

    def _SS22(self, ini_val, input_func, inputfunc_args):
        ''' Calculate the dynamic solution with SS22

        !NOT CALLABLE!
        Solves the second order ode using the SS22 algorithm [Zienkiewicz].

        Args:
        ini_val: Initial value for the solution x(t=t0)
        input_func: Function which describes the input over time
        inputfunc_args: Arguments for the input function
        '''

        self.ksp_tran.setFromOptions()

        xn = ini_val[0]
        dxn = ini_val[1]

        self.t_sol.append(0.0)
        self.x_sol.append(xn)

        alpha = self.K.getVecRight()

        for k in range(self.N):

            xmean = xn + self.solver_parameter[0] * self.dt * dxn
            dxmean = dxn

            tmp = input_func(k * self.dt, xn, inputfunc_args)
            rhs = -self.C * dxmean - self.K * xmean + self.B * tmp

            alpha.set(0)

            print('\t\t\tTimestep {}'.format(k + 1))

            self.ksp_tran.solve(rhs, alpha)

            xn = xn + self.dt * dxn + 0.5 * pow(self.dt, 2) * alpha
            dxn = dxn + self.dt * alpha

            self.x_sol.append(xn.copy())
            self.t_sol.append((k + 1) * self.dt)
