# Copyright (C) 2019  Chair of Automatic Control, Kiel University
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from firedrake import *
from firedrake.petsc import PETSc
import numpy as np
import scipy as sp
import scipy.integrate
import scipy.io as sio
import h5py
import os
import csv
import argparse

from mpi4py import MPI

from dyn_solver import *


parser = argparse.ArgumentParser(
    description='Choose between the step input and the sinusoidal input.')
parser.add_argument(
    '--sinus',
    default=False,
    action='store_true',
    help='Enable the sinusoidal input.')

options = parser.parse_args()

if options.sinus:

    def input_function(t, x, input_args):
        '''
        Time dependent input function for the dynamic solution!
        '''

        if t < 0.1:
            return 0
        else:
            return 2500 * np.sin(2.0 * np.pi * 5.0 *
                                 (t - 0.1 - 5.0 / 4.0)) + 2500

    simType = 'sin'

else:

    def input_function(t, x, input_args):
        '''
        Time dependent input function for the dynamic solution!
        '''

        if t < 0.1:
            return 0
        else:
            return 5E3
    simType = 'step'


listSysString = [
    'fish_ext_bt8', 'fish_ext_hybrid_flbt', 'fish_ext_hybrid_flbt5',
    'fish_ext_hybrid_flbt5', 'fish_ext_flbt8', 'fish_ext_mt',
    'fish_ext_moment_matching'
]
listSysSubSting = ['8_p', '_p', '5_p', '5_fo', '8_fv', '_mt', '_mm10']

if len(listSysString) is not len(listSysString):
    raise ImportError(
        'Number of system strings does not match number of sub system strings.'
    )

mesh = Mesh('./data/fish_mesh.e')
V = VectorFunctionSpace(mesh, 'CG', 2)

stepsize = 0.001
T = 2.0

if not os.path.isfile('./results/output_full_' + simType + '/full_sys.mat'):

    try:
        os.mkdir('results')
    except:
        pass

    try:
        os.mkdir('results/output_full_' + simType)
    except:
        pass

    try:
        os.mkdir('./results/output_full_' + simType + '/state')
    except:
        pass

    print('Load full system')

    f = h5py.File('./data/fish_tail.mat')
    fullMSp = sp.sparse.csc_matrix(
        (f['M']['data'], f['M']['ir'], f['M']['jc'])).tocsr()
    fullMPet = PETSc.Mat().createAIJ(
        size=fullMSp.shape, csr=(
            fullMSp.indptr, fullMSp.indices, fullMSp.data))

    fullKSp = sp.sparse.csc_matrix(
        (f['K']['data'], f['K']['ir'], f['K']['jc'])).tocsr()
    fullKPet = PETSc.Mat().createAIJ(
        size=fullKSp.shape, csr=(
            fullKSp.indptr, fullKSp.indices, fullKSp.data))

    fullCSp = sp.sparse.csc_matrix(
        (f['E']['data'], f['E']['ir'], f['E']['jc'])).tocsr()
    fullCPet = PETSc.Mat().createAIJ(
        size=fullCSp.shape, csr=(
            fullCSp.indptr, fullCSp.indices, fullCSp.data))

    fullFSp = f['Bu'].value.transpose()
    fullFPet = PETSc.Vec().createWithArray(fullFSp).setUp()
    fullFPet.assemblyBegin()
    fullFPet.assemblyEnd()

    fullCoutSp = sp.sparse.csr_matrix(f['Cp_ext'].value.transpose())

    fullSolve = dyn_solver(
        fullMPet,
        fullCPet,
        fullKPet,
        fullFPet,
        T,
        stepsize,
        system_string='full system')

    kspFull = fullSolve.ksp_tran
    kspFull.setType('preonly')
    pcFull = kspFull.getPC()
    pcFull.setType('lu')
    pcFull.setFactorSolverType('mumps')

    x0 = fullKPet.getVecRight()
    x0.set(0)
    dx0 = x0.copy()

    fullSolution = fullSolve.start_solve([x0, dx0], input_function)

    fullTime = fullSolution[0]
    fullState = []
    fullOutput = []
    fullInput = []

    for i in range(len(fullSolution[0])):
        fullState.append(fullSolution[1][i].array)
        fullOutput.append(fullCoutSp.dot(fullState[i]))
        fullInput.append(input_function(fullTime[i], [], []))

        with open(
                './results/output_full_' + simType +
            '/state/state' + str(i) + '.csv',
                mode='w') as writeState:
            stateWriter = csv.writer(writeState, delimiter=',')
            stateWriter.writerow(fullState[i])

    fullState = np.array(fullState)
    fullOutput = np.array(fullOutput)

    fullSaveDict = {}
    fullSaveDict['t'] = fullTime
    fullSaveDict['y'] = fullOutput

    sio.savemat('./results/output_full_' + simType + '/full_sys', fullSaveDict)

else:

    print(
        'Load full system simulation results - If you want to simulate the full system again delete the file ./output_full/full_sys.mat!'
    )
    fullLoadDict = sio.loadmat('./results/output_full_' + simType + '/full_sys')

    fullOutput = list(fullLoadDict['y'])
    fullTime = list(fullLoadDict['t'][0])

for r in range(len(listSysString)):

    redSysString = listSysString[r]
    redSysSubString = listSysSubSting[r]

    print('\nLoad reduced system matrices for ' + redSysString + ' using rom' +
          redSysSubString)

    redLoadDict = sio.loadmat('./data/' + redSysString + '.mat')

    try:
        redMNp = redLoadDict['rom' + redSysSubString]['M'][0, 0]
        redKNp = redLoadDict['rom' + redSysSubString]['K'][0, 0]
        redENp = redLoadDict['rom' + redSysSubString]['E'][0, 0]
        redBNp = redLoadDict['rom' + redSysSubString]['Bu'][0, 0]
        redCpNp = redLoadDict['rom' + redSysSubString]['Cp'][0, 0]

        try:
            redCvNp = redLoadDict['rom' + redSysSubString]['Cv'][0, 0]
        except:
            pass

        try:
            redTrNp = redLoadDict['info' + redSysSubString]['T'][0, 0]
            redTlNp = redLoadDict['info' + redSysSubString]['W'][0, 0]
        except:
            redTrNp = np.zeros([779232, redMNp.shape[0]])
            redTlNp = np.zeros([779232, redMNp.shape[0]])

        if (np.iscomplex(redMNp) == True).any():
            redMNp = np.real(redMNp)
            redKNp = np.real(redKNp)
            redENp = np.real(redENp)
            redBNp = np.real(redBNp)
            redCpNp = np.real(redCpNp)
            redTrNp = np.real(redTrNp)
            redTlNp = np.real(redTlNp)

        redSecOrderFlag = True
    except:
        try:
            redANp = redLoadDict['rom' + redSysSubString]['A'][0, 0]
            redBNp = redLoadDict['rom' + redSysSubString]['B'][0, 0]
            redCNp = redLoadDict['rom' + redSysSubString]['C'][0, 0]
            redDNp = redLoadDict['rom' + redSysSubString]['D'][0, 0]
            redENp = redLoadDict['rom' + redSysSubString]['E'][0, 0]
            try:
                redTrNp = redLoadDict['rom' + redSysSubString]['T'][0, 0]
                redTlNp = redLoadDict['rom' + redSysSubString]['W'][0, 0]
            except:
                redTrNp = np.zeros([779232, redANp.shape[0]])
                redTlNp = np.zeros([779232, redANp.shape[0]])

            redSecOrderFlag = False
        except:
            raise ValueError('Please enter a valid .mat File.')

    try:
        os.mkdir('./results/output_reduced_' + simType)
    except:
        pass

    try:
        os.mkdir('./results/output_reduced_' + simType + '/' + redSysString)
    except:
        pass

    if redSecOrderFlag:
        redMPet = PETSc.Mat().createAIJ((redMNp.shape)).setUp()
        redMPet.setValues(
            range(redMPet.size[0]), range(redMPet.size[1]), redMNp)
        redMPet.assemblyBegin()
        redMPet.assemblyEnd()

        redEPet = PETSc.Mat().createAIJ((redMNp.shape)).setUp()
        redEPet.setValues(
            range(redMPet.size[0]), range(redMPet.size[1]), redENp)
        redEPet.assemblyBegin()
        redEPet.assemblyEnd()

        redKPet = PETSc.Mat().createAIJ((redMNp.shape)).setUp()
        redKPet.setValues(
            range(redMPet.size[0]), range(redMPet.size[1]), redKNp)
        redKPet.assemblyBegin()
        redKPet.assemblyEnd()

        redBPet = PETSc.Vec().createWithArray(redBNp)

        x0 = redMPet.getVecRight()
        x0.set(0)
        dx0 = x0.copy()
        ddx0 = x0.copy()

        redSolver = dyn_solver(
            redMPet,
            redEPet,
            redKPet,
            redBPet,
            T,
            stepsize,
            system_string='reduced system ' + redSysString)

        kspRed = redSolver.ksp_tran
        kspRed.setType('preonly')
        pcRed = kspRed.getPC()
        pcRed.setType('lu')
        pcRed.setFactorSolverType('mumps')

        redSolution = redSolver.start_solve([x0, dx0], input_function)

        redFullState = []
        redState = []
        redOutput = []
        redTime = redSolution[0]

        for i in range(len(redSolution[0])):
            redState.append(redSolution[1][i].array)
            redFullState.append(redTrNp.dot(redSolution[1][i].array))
            redOutput.append(redCpNp.dot(redState[i]))

        redOut = np.array(redOutput)
        redFullState = np.array(redFullState)

    else:
        x0Red = np.array(np.zeros(redANp.shape)[0, :])

        def redRhs(x, t, A, b, inputfunc):
            return A.dot(x) - b * inputfunc(t, [], [])

        numsteps = int(T / stepsize) + 1
        t = np.linspace(0, T, num=numsteps)

        print('Solve reduced first order system ' + redSysString)
        redSolution = sp.integrate.odeint(
            redRhs, x0Red, t, args=(redANp, redBNp[:, 0], input_function))

        redTime = t
        redOutput = []
        redState = []
        redFullState = []

        for i in range(len(redTime)):
            redState.append(redSolution[i, :])
            redOutput.append(redCNp.dot(redState[i]))
            redFullState.append(redTrNp.dot(redState[i]))

        redOut = np.array(redOutput)
        redFullState = np.array(redFullState)

    print('\tSave Results')
    redSaveDict = {}
    redSaveDict['x'] = redState
    redSaveDict['t'] = redTime
    redSaveDict['y'] = redOutput

    sio.savemat(
        './results/output_reduced_' + simType + '/' + redSysString + '/' + redSysString + '_' +
        redSysSubString, redSaveDict)

    print('\tDetermine Errornorm (this may take some time)')
    errOutput = []
    xFull = Function(V)
    xRed = Function(V)
    sInput = []
    errNorm = []

    for i in range(len(fullTime)):
        print('\t\tStep {} of {}'.format(i, len(fullTime)))
        sInput.append(input_function(fullTime[i], [], []))
        errOutput.append(fullOutput[i] - redOutput[i])

        with open('./results/output_full_' + simType + '/state/state' + str(i) + '.csv') as csv_file:
            csv_reader = csv.reader(
                csv_file, delimiter=',', quoting=csv.QUOTE_NONNUMERIC)
            for row in csv_reader:
                tmpFullState = np.array(row)

        xFull.vector().set_local(tmpFullState)
        xRed.vector().set_local(redFullState[i, :])

        errNorm.append(errornorm(xFull, xRed))

    print('\tWrite Output')
    compSaveDict = {}
    compSaveDict['Time'] = fullTime
    compSaveDict['Input'] = sInput
    compSaveDict['yFull'] = fullOutput
    compSaveDict['yRed'] = redOutput
    compSaveDict['yErr'] = errOutput
    compSaveDict['errornorm_state'] = errNorm

    sio.savemat('./results/output_reduced_' + simType + '/' + redSysString + '/comp' + redSysSubString,
                compSaveDict)

    inputSaveDict = {}
    inputSaveDict['Time'] = fullTime
    inputSaveDict['Input'] = sInput

    sio.savemat(
        './results/output_reduced_' + simType + '/' +
        redSysString + '/input' + redSysSubString,
        compSaveDict)

    print('Calculations for ' + redSysString + ' using rom' + redSysSubString +
          ' completed')
